"""This module provides support for the unilateral inverse Laplace
transform.

Copyright 2021 Michael Hayes, UCECE

"""

from .transformer import UnilateralInverseTransformer
from .ratfun import Ratfun
from .sym import simplify, AppliedUndef
from .utils import factor_const, scale_shift, as_sum_terms
from .matrix import matrix_inverse
from warnings import warn
import sympy as sym

Zero = sym.S.Zero
One = sym.S.One

__all__ = ('ILT', 'inverse_laplace_transform')


class InverseLaplaceTransformer(UnilateralInverseTransformer):

    name = 'inverse Laplace transform'

    def noevaluate(self, expr, s, t):

        # Construct Bromwich integral.
        self.error('TODO')

    def check(self, expr, s, t, **kwargs):

        if expr.has(t):
            self.error('Expression depends on t')

    def key(self, expr, s, t, **kwargs):
        return (expr, s, t,
                kwargs.get('causal', False),
                kwargs.get('damped_sin', True),
                kwargs.get('damping', None))

    def func(self, expr, s, t):

        if not isinstance(expr, AppliedUndef):
            self.error('Expecting function')

        scale, shift = scale_shift(expr.args[0], s)

        # Convert V(s) to v(t), etc.
        name = expr.func.__name__
        undef = sym.Function(name[0].lower() + name[1:])(t)

        result = undef.subs(t, t / scale) / abs(scale)

        if shift != 0:
            result = result * sym.exp(t * shift / scale)
        return result

    def do_damped_sin(self, expr, s, t):

        ncoeffs, dcoeffs = expr.coeffs()
        K = ncoeffs[0] / dcoeffs[0]

        ncoeffs = [(c / ncoeffs[0]) for c in ncoeffs]
        dcoeffs = [(c / dcoeffs[0]) for c in dcoeffs]

        if len(ncoeffs) > 3 or len(dcoeffs) > 3:
            self.error('Not a second-order response')

        omega0 = sym.sqrt(dcoeffs[2])
        zeta = dcoeffs[1] / (2 * omega0)

        if zeta.is_constant() and zeta > 1:
            warn('Expression is overdamped')

        sigma1 = (zeta * omega0).simplify()
        omega1 = (omega0 * sym.sqrt(1 - zeta**2)).simplify()
        K = (K / omega1).simplify()

        E = sym.exp(-sigma1 * t)
        S = sym.sin(omega1 * t)

        h = K * E * S

        # If overdamped
        #h = K * sym.exp(-sigma1 * t) * sym.sinh(omega0 * mu * t)

        if len(ncoeffs) == 1:
            return Zero, h

        C = sym.cos(omega1 * t)
        kCd = omega1
        kSd = -sigma1
        hd = K * E * (kCd * C + kSd * S)

        if len(ncoeffs) == 2:
            return Zero, K * E * (kCd * C + (ncoeffs[1] + kSd) * S)

        kCdd = -2 * omega1 * sigma1
        kSdd = sigma1**2 - omega1**2

        G = K * E * ((kCdd + ncoeffs[1] * kCd) * C + (kSdd + ncoeffs[1] * kSd + ncoeffs[2]) * S)

        return K * kCd * sym.DiracDelta(t), G

    def ratfun(self, expr, s, t, **kwargs):

        if kwargs.pop('pdb', False):
            import pdb; pdb.set_trace()

        sexpr = Ratfun(expr, s)

        if kwargs.get('damped_sin', False):
            if sexpr.degree == 2:
                return self.do_damped_sin(sexpr, s, t)
            #if False and sexpr.degree == 3 and Ratfun(expr * s).degree == 2:
            #    return self.do_damped_sin3(sexpr, s, t)

        self.debug('Finding QRPO representation')

        damping = kwargs.get('damping', None)
        Q, R, P, O, delay, undef = sexpr.as_QRPO(damping)

        if delay != 0:
            # This will be caught and trigger expansion of the expression.
            raise ValueError('Unhandled delay %s' % delay)

        cresult = Zero

        if Q:
            Qpoly = sym.Poly(Q, s)
            C = Qpoly.all_coeffs()
            for n, c in enumerate(C):
                cresult += c * sym.diff(sym.DiracDelta(t), t, len(C) - n - 1)

        if R == []:
            return cresult, 0

        uresult = 0
        for m, (p, n, A) in enumerate(zip(P, O, R)):

            # This is zero for the conjugate pole.
            if R[m] == 0:
                continue

            # Search and remove conjugate pair.
            has_conjpair = False
            if p.is_complex and kwargs.get('pairs', True):
                pc = p.conjugate()
                Ac = A.conjugate()
                for m2, p2 in enumerate(P[m + 1:]):
                    m2 += m + 1
                    if n == O[m2] and p2 == pc and R[m2] == Ac:
                        R[m2] = 0
                        has_conjpair = True
                        break

            if has_conjpair:
                # Combine conjugate pairs.
                p = p.expand(complex=True)
                A = A.expand(complex=True)
                p_re = sym.re(p)
                p_im = sym.im(p)
                A_re = sym.re(A)
                A_im = sym.im(A)
                et = sym.exp(p_re * t)
                result = 2 * A_re * et * sym.cos(p_im * t)
                result -= 2 * A_im * et * sym.sin(p_im * t)
            else:
                result = A * sym.exp(p * t)

            if n > 1:
                result *= t ** (n - 1) / sym.factorial(n - 1)
            uresult += result

        # cresult is a sum of Dirac deltas and its derivatives so is known
        # to be causal.

        return cresult, uresult

    def product(self, expr, s, t, **kwargs):

        # Handle expressions with a function of s, e.g., V(s) * Y(s), V(s)
        # / s etc.
        if kwargs.get('causal', False):
            # Assume that all functions are causal in the expression.
            t1 = Zero
            t2 = t
        else:
            t1 = -sym.oo
            t2 = sym.oo

        const, expr = factor_const(expr, s)

        factors = expr.as_ordered_factors()
        if len(factors) < 2:
            cresult, uresult = self.term1(expr, s, t, **kwargs)
            return const * (cresult + uresult)

        if (len(factors) > 2 and not
            # Help s * 1 / (s + R * C) * I(s)
            isinstance(factors[1], AppliedUndef) and
            isinstance(factors[2], AppliedUndef)):
            factors = [factors[0], factors[2], factors[1]] + factors[3:]

        if isinstance(factors[1], AppliedUndef):
            # Try to expose more simple cases, e.g. (R + s * L) * V(s)
            terms = factors[0].expand().as_ordered_terms()
            if len(terms) >= 2:
                result = Zero
                for term in terms:
                    result += self.product(factors[1] * term, s, t, **kwargs)
                return result * const

        cresult, uresult = self.term1(factors[0], s, t, **kwargs)
        result = cresult + uresult

        intnum = 0
        for m in range(len(factors) - 1):
            if m == 0 and isinstance(factors[1], AppliedUndef):
                # Note, as_ordered_factors puts powers of s before the functions.
                if factors[0] == s:
                    # Handle differentiation
                    # Convert s * V(s) to d v(t) / dt
                    result = self.func(factors[1], s, t)
                    result = sym.Derivative(result, t)
                    continue
                elif factors[0].is_Pow and factors[0].args[0] == s and factors[0].args[1] > 0:
                    # Handle higher order differentiation
                    # Convert s ** 2 * V(s) to d^2 v(t) / dt^2
                    result = self.func(factors[1], s, t)
                    result = sym.Derivative(result, t, factors[0].args[1])
                    continue
                elif factors[0].is_Pow and factors[0].args[0] == s and factors[0].args[1] == -1:
                    # Handle integration  1 / s * V(s)
                    tau = self.dummy_var(expr, 'tau', level=intnum, real=True)
                    intnum += 1
                    result = self.func(factors[1], s, tau)
                    result = sym.Integral(result, (tau, t1, t))
                    continue
            # Convert product to convolution
            tau = self.dummy_var(expr, 'tau', level=intnum, real=True)
            intnum += 1
            cresult, uresult = self.term1(factors[m + 1], s, t, **kwargs)
            expr2 = cresult + uresult
            result = sym.Integral(result.subs(t, t - tau) * expr2.subs(t, tau),
                                  (tau, t1, t2))

        return result * const

    def power(self, expr, s, t):

        # Handle expressions with a power of s.
        if not (expr.is_Pow and expr.args[0] == s):
            self.error('Expression not a power of s')
        exponent = expr.args[1]

        # Have many possible forms; the common ones are:
        # s**a, s**-a, s**(1+a), s**(1-a), s**-(1+a), s**(a-1)
        # Cannot tell if 1-a is positive.

        if exponent.is_positive:
            # Unfortunately, SymPy does not seem to support fractional
            # derivatives...
            return sym.Derivative(sym.DiracDelta(t), t, exponent,
                                  evaluate=False)

        if exponent.is_negative:
            return sym.Pow(t, -exponent - 1) / sym.Gamma(-exponent)

        self.error('Cannot determine sign of exponent')

    def delay_factor(self, expr, var):

        delay = Zero
        rest = One

        for f in expr.as_ordered_factors():
            b, e = f.as_base_exp()
            if b == sym.E and e.is_polynomial(var):
                p = sym.Poly(e, var)
                c = p.all_coeffs()
                if p.degree() == 1:
                    delay -= c[0]
                    if c[1] != 0:
                        rest *= sym.exp(c[1])
                    continue

            rest *= f
        return rest, delay

    def sympy(self, expr, s, t):

        self.debug('Resorting to SymPy')

        # This barfs when needing to generate Dirac deltas
        from sympy.integrals.transforms import inverse_laplace_transform
        result = inverse_laplace_transform(expr, s, t)

        if result.has(sym.InverseLaplaceTransform):
            raise ValueError('SymPy cannot find inverse Laplace transform of %s' % expr)
        return result

    def term1(self, expr, s, t, **kwargs):

        const, expr = factor_const(expr, s)

        if isinstance(expr, AppliedUndef):
            # Handle V(s), 3 * V(s) etc.  If causal is True it is assumed
            # that the unknown functions are causal.
            result = self.func(expr, s, t)
            return result * const, Zero

        if expr.has(AppliedUndef):
            return const * self.product(expr, s, t, **kwargs), Zero

        try:
            # This is the common case.
            cresult, uresult = self.ratfun(expr, s, t, **kwargs)
            return const * cresult, const * uresult
        except:
            pass

        if expr.is_Pow and expr.args[0] == s:
            return Zero, const * self.power(expr, s, t)

        raise ValueError('Cannot determine inverse Laplace transform')

    def term(self, expr, s, t, **kwargs):

        expr, delay = self.delay_factor(expr, s)

        if delay == 0 and expr.has(sym.exp):
            # Handle cases like 1 / (s**2 * exp(5 * s) + s * exp(5 * s))
            expr1 = expr.simplify()
            expr2, delay2 = self.delay_factor(expr1, s)
            if not expr2.has(sym.exp):
                # Simplify can make things worse, e.g., 1 - exp(-5 *s)
                # becomes exp(-5 * s) * (exp(5 * s) - 1)
                expr = expr2
                delay = delay2

        try:
            cresult, uresult = self.term1(expr, s, t, **kwargs)
        except:

            # With deep=True, SymPy makes mess of (1 - exp(-s * T))
            terms = expr.expand(deep=False).as_ordered_terms()

            if len(terms) > 1:

                try:
                    # See if can convert to convolutions...
                    return self.product(expr, s, t, **kwargs), Zero
                except:
                    pass

                uresult = Zero
                cresult = Zero

                for term in terms:
                    term = term.simplify()
                    cterm, uterm = self.term(term, s, t, **kwargs)
                    cresult += cterm
                    uresult += uterm
                return cresult, uresult

            expr = expr.simplify()

            try:
                cresult, uresult = self.term1(expr, s, t, **kwargs)
            except:
                return Zero, self.sympy(expr, s, t)

        if delay != 0:
            cresult = cresult.subs(t, t - delay)
            uresult = uresult.subs(t, t - delay)

            # h(t) = g(t - T)
            # If h(t) is known to be causal and T >= 0, then g(t)
            # is also causal.  If T > 0, then causality is violated.

            # If h(t) is not known to be causal then we can only infer
            # it for t >= 0 from its Laplace transform H(s).
            # From the delay theorem, H(s) = G(s) * exp(-s * T).
            # So given G(s) we can only infer g(t) for t >= 0.
            # This implies that we can only infer h(t) for t >= T.
            # This creates a can of worms since different terms of
            # an expression can have different conditions, say
            # exp(-3 * s) / s**2 + exp(-4 * s) / s.  So for now,
            # assume causal expression...

            if not kwargs.get('causal', False) and expr.has(AppliedUndef):
                warn('Assuming causal expression')

            if not delay.is_negative:
                if not delay.is_positive:
                    warn('Assuming %s is positive' % delay)
                cresult += uresult * sym.Heaviside(t - delay)
                uresult = Zero
            else:
                raise ValueError('Causality violated with time advance %s.' % delay)

        else:
            if kwargs.get('causal', False):
                cresult += uresult * sym.Heaviside(t)
                uresult = Zero

        return cresult, uresult


inverse_laplace_transformer = InverseLaplaceTransformer()


def inverse_laplace_transform(expr, s, t, evaluate=True, **kwargs):
    """Calculate inverse Laplace of X(s) and return x(t).

    The unilateral Laplace transform cannot determine x(t) for n < 0
    unless given additional information:

    `dc` -- x(t) = constant
    `causal` -- x(t) = 0 for n < 0.
    `ac` -- x(t) = A cos(a * n) + B * sin(b * n)
    """

    return inverse_laplace_transformer.transform(expr, s, t,
                                                 evaluate=evaluate,
                                                 **kwargs)


def ILT(expr, s, t, evaluate=True, **kwargs):
    """Calculate inverse Laplace of X(s) and return x(t).

    The unilateral Laplace transform cannot determine x(t) for n < 0
    unless given additional information:

    `dc` -- x(t) = constant
    `causal` -- x(t) = 0 for n < 0.
    `ac` -- x(t) = A cos(a * n) + B * sin(b * n)
    """

    return inverse_laplace_transformer.transform(expr, s, t,
                                                 evaluate=evaluate,
                                                 **kwargs)



"""This module provides the NetlistHelper class used by the
NetlistMaker and LadderMaker classes.

Copyright 2020--2021 Michael Hayes, UCECE

"""

from .componentnamer import ComponentNamer


class NetlistHelper(object):

    evalf = False
    
    @property 
    def _node(self):

        if not hasattr(self, '_node_counter'):
            self._node_counter = 0
        ret = self._node_counter
        self._node_counter += 1
        return ret

    def _make_nodes(self, *nodes):

        return [self._node if node is None else node for node in nodes]
    
    def _make_name(self, kind):
        """Make identifier"""

        if not hasattr(self, '_namer'):
            self._namer = ComponentNamer()

        return self._namer.name(kind, '')

    def _make_id(self, kind):
        """Make identifier"""

        # Perhaps the names should try to follow the value names, say
        # when generated by the random network generator?  This will
        # avoid confusion.
        if not hasattr(self, '_namer'):
            self._namer = ComponentNamer()

        return self._namer.netid(kind, '')    

    def _netarg(self, arg):

        if self.evalf:
            arg = arg.evalf(n=self.evalf)

        arg = str(arg)

        # TODO: make more robust to catch expressions.
        if ('(' in arg) or (')' in arg) or (' ' in arg) or (',' in arg) or ('*' in arg) or ('/' in arg):
            return '{%s}' % arg
        return arg

    def _netargs(self, net):

        return ' '.join([self._netarg(arg) for arg in net.args])




"""This module provides discrete-time filter support.

Copyright 2021 Michael Hayes, UCECE

"""

from .expr import expr, equation, ExprTuple
from .nexpr import DiscreteTimeDomainExpression
from .diffeq import DifferenceEquation
from .discretetime import n, z, seq
from .sequence import Sequence
from .utils import isiterable
from numpy import arange, ndarray
from scipy.signal import lfilter
import sympy as sym


class DLTIFilter(object):

    def __init__(self, b, a):
        """Create discrete-time filter where `b` is a list or array of
        numerator coefficients and `a` is a list or array of
        denominator coefficients."""

        if not isiterable(b):
            b = (b, )
        if not isiterable(a):
            a = (a, )            
        
        self.a = ExprTuple(a)
        self.b = ExprTuple(b)

    def __repr__(self):
        """This is called by repr(expr).  It is used, e.g., when printing
        in the debugger."""
        
        return '%s(%s, %s)' % (self.__class__.__name__, self.b, self.a)

    def transfer_function(self):
        """Return discrete-time impulse response (transfer function) in
        z-domain."""                

        Nl = len(self.a)
        Nr = len(self.b)
        
        # numerator of H(z)
        num = 0 * z 
        for i in range(Nr): 
            num += self.b[i] * z**(-i)
            
        # denominator for H(z)
        denom = self.a[0] * z**0  
        for k in range(1, Nl):
            az = self.a[k] * z**(-k)
            denom += az
  
        # collect with respect to positive powers of the variable z
        num = sym.collect(sym.expand(num * z**Nl), z)
        denom = sym.collect(sym.expand(denom * z**Nl), z)
        
        Hz = expr(sym.simplify(num / denom))
        Hz.is_causal = True
        return Hz

    def impulse_response(self):
        """Return discrete-time impulse response (transfer function) in time domain."""        

        H = self.transfer_function()
        return H(n)

    def difference_equation(self, inputsym='x', outputsym='y'):
        """Return difference equation."""

        rhs = 0 * n

        for m, bn in enumerate(self.b):
            rhs += bn * expr('%s(n - %d)' % (inputsym, m))

        for m, an in enumerate(self.a[1:]):
            rhs -= an * expr('%s(n - %d)' % (outputsym, m + 1))

        lhs = self.a[0] * expr('y(n)')
        return DifferenceEquation(lhs, rhs, inputsym, outputsym)

    def zdomain_initial_response(self, ic):
        """Return Z-domain response due to initial conditions.
           ic : list with initial values y[-1], y[-2], ... 
        """        

        Nl = len(self.a)
        
        # Denominator for Yi(z)
        denom = self.a[0] * z**0  
        num = 0 * z
        for k in range(1, Nl):
            az = self.a[k] * z**(-k)
            denom += az
            # Numerator for Yi(z)
            y0 = 0 * z
            for i in range(0, k):
                y0 += ic[i] * z**(i + 1)
            num += az * y0
  
        # Collect with respect to positive powers of the variable z
        num = sym.collect(sym.expand(num * z**Nl), z)
        denom = sym.collect(sym.expand(denom * z**Nl), z)
  
        Yzi = expr(-sym.simplify(num / denom))
        Yzi.is_causal = True
        
        return Yzi

    def initial_response(self, ic):
        """Return response due to initial conditions in the time domain.
           ic : list with initial values y[-1], y[-2], ...
        """

        Yzi = self.zdomain_initial_response(ic)
        return Yzi(n)
    
    def response(self, x, ic=None, ni=None):
        """Calculate response of filter to input `x` given a list of initial conditions
        `ic` for time indexes specified by `ni`.  If `ni` is a tuple,
        this specifies the first and last (inclusive) time index. 
        
        The initial conditions are valid prior to the time indices given by the ni
        `x` can be an expression, a sequence, or a list/array of values.
        """

        if (ic is None and ni is None and isinstance(x, ndarray)
            and self.a.symbols == {} and self.b.symbols == {}):

            # Calculate numerical response
            b = self.b.fval
            a = self.a.fval
            return lfilter(b, a, x)

        if ic is None:
            ic = [0] * (len(self.a) - 1)

        if not isiterable(x) and not isinstance(x, DiscreteTimeDomainExpression):
            x = (x, )

        if not isiterable(ic):
            ic = (ic, )            
        
        if isinstance(x, (tuple, list, ndarray)):
            x = seq(x)
        elif not isinstance(x, (Sequence, DiscreteTimeDomainExpression)):
            raise ValueError('The input x must be a scalar, tuple, sequence, nexpr, list, or array')

        NO = len(ic)
        
        if NO != len(self.a) - 1:
            raise ValueError("Expected %d initial conditions, got %d" % (len(self.a) - 1, NO))

        if ni is None:
            ni = (0, 10)
        
        if isinstance(ni, tuple):
            ni = arange(ni[0], ni[1] + 1)
        
        Nn = len(ni)
  
        # Order right hand side
        Nr = len(self.b)
  
        y_tot = list(ic[-1::-1]) + Nn * [0]
  
        a_r = self.a[-1:-1-NO:-1]
        for i, nval in enumerate(ni):
            # Get previous y vals (sliding window)
            pre_y = y_tot[i:i + NO]
    
            # Calculate rhs of new value
            if isinstance(x, Sequence):
                rhs = sum(self.b[l] * x[nval - l] for l in range(Nr))
            else:
                rhs = sum(self.b[l] * x(nval - l) for l in range(Nr))
    
            # Add lhs
            y_tot[i + NO] = -1 / self.a[0] * sum(csi * ysi for csi, ysi in zip(a_r, pre_y)) + rhs / self.a[0]

        # Solution, without initial values
        ret_seq = seq(y_tot[NO:], ni)  
  
        return ret_seq


    def subs(self, *args, **kwargs):

        a = self.a.subs(*args, **kwargs)
        b = self.b.subs(*args, **kwargs)

        return self.__class__(b, a)



"""This module implements the DiscreteTimeDomainMatrix class for a matrix of
time-domain expressions.

Copyright 2021 Michael Hayes, UCECE

"""

from .matrix import Matrix

class DiscreteTimeDomainMatrix(Matrix):
    from .nexpr import DiscreteTimeDomainExpression    
    _typewrap = DiscreteTimeDomainExpression    

    def ZT(self):

        def func(expr):
            return expr.ZT()
        
        return ZDomainMatrix(self.applyfunc(func))
    
from .zmatrix import ZDomainMatrix



"""
This module implements the Lcapy Matrix class.

Copyright 2019--2021 Michael Hayes, UCECE
"""

import sympy as sym
from copy import copy
from .expr import expr
from .sym import simplify
from .printing import pprint, latex, pretty


def msympify(expr):
    # If do nothing, will get a problem with matrices that
    # have mixed data types, e.g., A matrix.

    if isinstance(expr, Expr):
        # Bye bye Lcapy type information...
        return expr.sympy
    return sym.sympify(expr)


class Matrix(sym.Matrix):

    # Unlike numpy.ndarray, the sympy.Matrix runs all the elements
    # through _sympify, creating sympy objects and thus losing the
    # original type information and associated methods.  As a hack, we
    # try to wrap elements when they are read using __getitem__.  This
    # assumes that all the elements have the same type.  This is not
    # the case for A, B, G, and H two-port matrices.  This could be
    # handled by having another matrix to specify the type for each
    # element.  What's worse, is that calling _sympify on each element
    # creates different variables than what we are expecting.  For
    # example, the LaplaceDomainExpression s looks the same but gets
    # different attributes.  We prevent this by defining _sympify.

    # Unfortunately, as of SymPy-1.9, having non-SymPy quantities
    # for SymPy matrix elements is deprecated.   Having Lcapy quantities
    # as SymPy matrix elements almost works for many operations.

    # An alternative approach is for Lcapy to define its own Matrix
    # class, similar to the SymPy Matrix class but independent.  This
    # would store Lcapy quantities for each matrix element.  The
    # advantage is that different matrix classes are not required for
    # each domain.  The down-sides are that there is a divergence with
    # SymPy matrix behaviour and that many of the SymPy Matrix
    # operations need to be reimplemented or wrapped.  Wrapping is
    # easier but the units may be wrong, say with a matrix inverse
    # operation.

    # _sympify is called for each matrix element when a matrix is created.
    _sympify = staticmethod(msympify)

    def __getitem__(self, key):

        # This wraps elements as an Expr.  It is called
        # for each element when a Matrix is printed.

        item = super(Matrix, self).__getitem__(key)

        # The following line is to handle slicing used
        # by the latex method.
        if isinstance(item, sym.Matrix):
            return item

        if hasattr(self, '_typewrap'):
            return self._typewrap(item)

        return expr(item, rational=False)

    def _repr_pretty_(self, p, cycle):
        """This is used by jupyter notebooks to display an expression using
        unicode."""

        p.text(pretty(self))

    def _repr_latex_(self):
        """This is used by jupyter notebooks to display an expression using
        LaTeX markup.  However, this requires mathjax.  If this method
        is not defined, jupyter falls back on _repr__pretty_ which
        outputs unicode."""
        return '$$' + latex(self) + '$$'

    def pprint(self):

        return pprint(self)

    def latex(self, **kwargs):

        return latex(self, **kwargs)

    def pdb(self):
        """Enter the python debugger."""

        import pdb; pdb.set_trace()
        return self

    def canonical(self):

        return self

    # TODO. There is probably a cunning way to automatically handle
    # the following.

    def inv(self, method='default'):
        Minv = matrix_inverse(sym.Matrix(self), method=method)

        return self.__class__(Minv)

    def det(self):

        return expr(super(Matrix, self).det())

    def norm(self):

        return expr(super(Matrix, self).norm())

    # TODO, either need to explicitly wrap methods or use some cunning
    # implicit method.

    def replace(self, query, value, map=False, simultaneous=True, exact=None):

        try:
            query = query.sympy
        except:
            pass

        try:
            value = value.sympy
        except:
            pass

        ret = super(Matrix, self).replace(query, value, map, simultaneous, exact)
        return self.__class__(ret)

    def rewrite(self, *args, **hints):

        f = lambda x: expr(x).rewrite(*args, **hints).sympy
        return self.applyfunc(f)

    def simplify(self):

        return self.applyfunc(simplify)

    def subs(self, *args, **kwargs):
        """Substitute variables in expression, see sympy.subs for usage."""

        f = lambda x: expr(x).subs(*args, **kwargs).sympy
        return self.applyfunc(f)

    @property
    def conj(self):
        """Complex conjugate; for compatilibility with Expr conj is an attribute."""
        return self._new(self.rows, self.cols, [x.conj for x in self])

    @property
    def symbols(self):

        symbols = {}
        for elt in self:
            symbols.update(expr(elt).symbols)
        return symbols

    @property
    def sympy(self):

        return sym.Matrix(self.rows, self.cols, [x.sympy for x in self])

    @property
    def expr(self):
        return self.sympy

    @property
    def is_complex(self):

        for x in self:
            if x.is_complex:
                return True
        return False

    def evaluate(self, arg=None):
        """Evaluate matrix at arg.  `arg` may be a scalar.
        The result is a NumPy float or complex array.

        There can be only one or fewer undefined variables in the expression.
        This is replaced by `arg` and then evaluated to obtain a result.
        """

        from numpy import empty

        dtype = float
        if self.is_complex:
            dtype = complex
        result = empty((self.rows, self.cols), dtype=dtype)

        for i in range(self.rows):
            for j in range(self.cols):
                result[i, j] = self[i, j].evaluate(arg)
        return result

    @property
    def numpy(self):
        """Return NumPy array; not a NumPy matrix."""
        return self.evaluate()


def matrix(mat):
    """Create Lcapy Matrix from a SymPy Matrix.

    If a t symbol is found in an element a tMatrix object is created.
    If a s symbol is found in an element an sMatrix object is created.

    """

    from .sym import tsym, ssym
    from .smatrix import LaplaceDomainMatrix
    from .tmatrix import TimeDomainMatrix

    elt = mat[0]
    try:
        elt = elt[0]
    except:
        pass

    if elt.has(tsym):
        return TimeDomainMatrix(mat)
    elif elt.has(ssym):
        return LaplaceDomainMatrix(mat)
    else:
        return mat


def matrix_inverse(M, method='default'):

    from .config import matrix_inverse_method, matrix_inverse_fallback_method

    if method == 'default':
        method = matrix_inverse_method

    if method == 'GE':
        try:
            from sympy.matrices import dotprodsimp

            # GE loses it without this assumption.  Well with
            # sympy-1.6.2 and the master version, GE still loses it
            # with a poor pivot.
            with dotprodsimp(False):
                return M.inv(method='GE')
        except:
            return M.inv(method='GE')

    elif method.startswith('DM-'):
        try:
            # This is experimental and requires sympy to be built from
            # git.  It only works for rational function fields but
            # fails for polynomial rings.  The latter can be handled
            # by converting it to a field, however, we just fall back
            # on a standard method.
            from sympy.polys.domainmatrix import DomainMatrix
            dM = DomainMatrix.from_list_sympy(*M.shape, rows=M.tolist())
            return dM.inv(method=method[3:]).to_Matrix()
        except:
            method = matrix_inverse_fallback_method

    return M.inv(method=method)

    def canonical(self):

        return self.applyfunc(self._typewrap.canonical)

    def general(self):

        return self.applyfunc(self._typewrap.general)

    def mixedfrac(self):

        return self.applyfunc(self._typewrap.mixedfrac)

    def partfrac(self):

        return self.applyfunc(self._typewrap.partfrac)

    def timeconst(self):

        return self.applyfunc(self._typewrap.timeconst)

    def ZPK(self):

        return self.applyfunc(self._typewrap.ZPK)


from .expr import Expr, expr



"""This module implements the LaplaceDomainMatrix class for a matrix of
Laplace-domain expressions.

Copyright 2019--2021 Michael Hayes, UCECE

"""

from .matrix import Matrix

class LaplaceDomainMatrix(Matrix):
    from .sexpr import LaplaceDomainExpression    
    _typewrap = LaplaceDomainExpression

    def ILT(self, **assumptions):

        def func(expr):
            return expr.ILT(**assumptions)
        
        return TimeDomainMatrix(self.applyfunc(func))


class LaplaceDomainVoltageMatrix(LaplaceDomainMatrix):
    from .sexpr import LaplaceDomainVoltage    
    _typewrap = LaplaceDomainVoltage

    
class LaplaceDomainCurrentMatrix(LaplaceDomainMatrix):
    from .sexpr import LaplaceDomainCurrent    
    _typewrap = LaplaceDomainCurrent

    
class LaplaceDomainAdmittanceMatrix(LaplaceDomainMatrix):
    from .sexpr import LaplaceDomainAdmittance    
    _typewrap = LaplaceDomainAdmittance

    
class LaplaceDomainImpedanceMatrix(LaplaceDomainMatrix):
    from .sexpr import LaplaceDomainImpedance    
    _typewrap = LaplaceDomainImpedance    
    
    
from .tmatrix import TimeDomainMatrix
    




"""This module wraps SymPy functions and provides a few others.
The underlying SymPy function is obtained with the sympy attribute.

Copyright 2014--2021 Michael Hayes, UCECE
"""

from .units import u as uu
import sympy as sym

__all__ = ('conjugate', 'sqrt', 'exp', 'log', 'log10', 'sin', 'cos',
           'tan', 'cot', 'asin', 'acos', 'atan', 'atan2', 'acot',
           'sinh', 'cosh', 'tanh', 'coth', 'asinh', 'acosh', 'atanh',
           'acoth', 'gcd', 'sign', 'diff', 'integrate', 'u', 'H',
           'heaviside', 'Heaviside', 'delta', 'DiracDelta', 'ui',
           'us', 'unitimpulse', 'unitstep', 'UnitImpulse', 'UnitStep',
           'rect', 'sinc', 'sincn', 'sincu', 'psinc', 'tri', 'trap',
           'Sum', 'dtrect', 'dtsign', 'Piecewise', 'Eq', 'Ne',
           'Lt', 'Le', 'Gt', 'Ge', 'Derivative', 'Integral', 'Max', 'Min',
           're', 'im')

from .extrafunctions import UnitImpulse as UnitImpulse1
from .extrafunctions import UnitStep as UnitStep1
from .extrafunctions import sincn as sincn1
from .extrafunctions import sincu as sincu1
from .extrafunctions import psinc as psinc1
from .extrafunctions import rect as rect1
from .extrafunctions import dtrect as dtrect1
from .extrafunctions import dtsign as dtsign1
from .extrafunctions import tri as tri1
from .extrafunctions import trap as trap1

class Function(object):

    def __init__(self, arg):

        if isinstance(arg, str):
            # Handled AppliedUndef
            arg = sym.Function(arg)
        self.expr = arg
    
    @property
    def sympy(self):
        """Return SymPy expression."""
        
        return self.expr

    def __call__(self, *args, **kwargs):

        func = self.expr
            
        e_args = list(map(expr, args))

        if func == sym.Piecewise:
            cls = e_args[0][0].__class__
        elif func in (sym.Integral, sym.integrate):
            # Integral(const, (tau, 0, t)) -> use t class
            # Integral(integrand, (tau, a, b)) -> use integrand class
            if e_args[0].is_constant:
                cls = e_args[1][2].__class__                
            else:
                cls = e_args[0].__class__            
        else:
            cls = e_args[0].__class__
            
            # Handle cases like atan2(1, omega)
            if len(e_args) > 1:
                if e_args[0].is_constant:
                    cls = e_args[1].__class__                

        try:
            if cls.is_discrete_time_domain or cls.is_discrete_fourier_domain:

                # TODO: fix for Heaviside change
                for old, new in function_mapping.items():
                    if func is old:
                        func = new
                        break
        except:
            pass
                
        result = func(*delcapify(e_args), **kwargs)

        result = cls(result)        
            
        if (self.expr in (sym.sin, sym.cos, sym.tan, sym.cot, sym.exp,
                          sym.sinh, sym.cosh, sym.tanh, sym.coth)
            and e_args[0].is_phase):
            result.part = ''
        elif self.expr in (sym.atan, sym.atan2):
            result.units = uu.rad
        elif self.expr == sym.diff and isinstance(e_args[1], Expr):
            result.units = e_args[0].units / e_args[1].units
        elif self.expr == sym.integrate:
            if isinstance(e_args[1], Expr):
                result.units = e_args[0].units * e_args[1].units
            elif isinstance(e_args[1], tuple) and isinstance(e_args[1][0], Expr):
                result.units = e_args[0].units * e_args[1][0].units
        elif self.expr is sym.DiracDelta and isinstance(e_args[0], Expr):
                result.units = 1 / e_args[0].units                

        return result

    def pdb(self):
        import pdb; pdb.set_trace()
        return self
    
    
class Log10(Function):

    # TODO, figure out how to print as log10(x) rather than
    # the expansion log(x) / log(10).  This will require
    # deferment of the expansion.
    
    def __call__(self, arg):
        return super(Log10, self).__call__(arg, 10)


class SincnFunction(Function):
    """Normalized sinc function :math:`\sin(\pi x)/(\pi x)`."""

    
class SincuFunction(Function):
    """Unnormalized sinc function :math:`\sin(x)/(x)`."""

    
class PsincFunction(Function):
    """Periodic sinc function :math:`\sin(M * x)/(M * sin(x))`."""        

    
conjugate = Function(sym.conjugate)

sqrt = Function(sym.sqrt)

exp = Function(sym.exp)

log = Function(sym.log)

log10 = Log10(sym.log)

sin = Function(sym.sin)

cos = Function(sym.cos)

tan = Function(sym.tan)

cot = Function(sym.cot)

asin = Function(sym.asin)

acos = Function(sym.acos)

atan = Function(sym.atan)

atan2 = Function(sym.atan2)

acot = Function(sym.acot)

sinh = Function(sym.sinh)

cosh = Function(sym.cosh)

tanh = Function(sym.tanh)

coth = Function(sym.coth)

asinh = Function(sym.asinh)

acosh = Function(sym.acosh)

atanh = Function(sym.atanh)

acoth = Function(sym.acoth)

gcd = Function(sym.gcd)

sign = Function(sym.sign)

diff = Function(sym.diff)

integrate = Function(sym.integrate)

u = H = heaviside = Heaviside = Function(sym.Heaviside)

delta = DiracDelta = Function(sym.DiracDelta)

Piecewise = Function(sym.Piecewise)

Derivative = Function(sym.Derivative)

Integral = Function(sym.Integral)

Eq = Function(sym.Eq)

Ne = Function(sym.Ne)

Lt = Function(sym.Lt)

Le = Function(sym.Le)

Gt = Function(sym.Gt)

Ge = Function(sym.Ge)

Add = Function(sym.Add)

Mul = Function(sym.Mul)

MatAdd = Function(sym.MatAdd)

MatMul = Function(sym.MatMul)

Sum = Function(sym.Sum)

Min = Function(sym.Min)

Max = Function(sym.Max)

re = Function(sym.re)

im = Function(sym.im)

ui = unitimpulse = UnitImpulse = Function(UnitImpulse1)

us = unitstep = UnitStep = Function(UnitStep1)

rect = Function(rect1)

dtrect = Function(dtrect1)

dtsign = Function(dtsign1)

sinc = SincnFunction(sincn1)

sincn = SincnFunction(sincn1)

sincu = SincuFunction(sincu1)

psinc = PsincFunction(psinc1)

tri = Function(tri1)

trap = Function(trap1)

function_mapping = {sym.Heaviside: UnitStep1,
                    sym.DiracDelta: UnitImpulse1,
                    rect1: dtrect,
                    sym.sign: dtsign}

from .expr import Expr, expr, delcapify



"""This module provides support for Laplace transforms.  It acts as a
quantity for SymPy's Laplace transform.  It calculates the unilateral
Laplace transform using:

F(s) = lim_{t_0\rightarrow 0} \int_{-t_0}^{\infty} f(t) e^{-s t} dt

In comparison, SymPy uses:

F(s) = \int_{0}^{\infty} f(t) e^{-s t} dt

The latter gives 0.5 for the Laplace transform of DiracDelta(t)
whereas the former version gives 1.  Note, SymPy is inconsistent in
that it gives DiracDelta(t) for the inverse Laplace transform of 1.

Another difference with this implementation is that it will transform
undefined functions such as v(t) to V(s).

These functions are for internal use by Lcapy.

Copyright 2016--2021 Michael Hayes, UCECE

"""

from .transformer import UnilateralForwardTransformer
from .ratfun import Ratfun
from .sym import sympify, simplify, AppliedUndef
from .utils import factor_const, scale_shift, as_sum_terms
import sympy as sym

__all__ = ('LT', 'laplace_transform')

class LaplaceTransformer(UnilateralForwardTransformer):

    name = 'Laplace transform'

    def noevaluate(self, expr, t, s):

        t0 = sympify('t0')
        return sym.Limit(sym.Integral(expr * sym.exp(-s * t),
                                      (t, t0, sym.oo)),
                         t0, 0, dir='-')

    def check(self, expr, t, s, **kwargs):

        if expr.has(s):
            self.error('Expression depends on s')

    def key(self, expr, t, s, **kwargs):
        return expr, t, s,

    def limits(self, expr, t, s, tmin, tmax):

        F = sym.integrate(expr * sym.exp(-s * t), (t, tmin, tmax))

        if not F.has(sym.Integral):
            return F

        if not F.is_Piecewise:
            self.error('Expecting piecewise')

        F, cond = F.args[0]
        if F.has(sym.Integral):
            self.error('Expecting integral')

        return F

    def do_0minus(self, expr, t, s):

        t0 = sym.symbols('t0', negative=True, real=True)

        F = self.limits(expr, t, s, t0, sym.oo)
        return sym.limit(F, t0, 0)

    def do_0(self, expr, t, s):

        return self.limits(expr, t, s, 0, sym.oo)

    def func(self, expr, t, s, inverse=False):

        if not isinstance(expr, AppliedUndef):
            self.error('Expecting function')

        scale, shift = scale_shift(expr.args[0], t)

        ssym = sympify(str(s))

        # Convert v(t) to V(s), etc.
        name = expr.func.__name__
        func = name[0].upper() + name[1:] + '(%s)' % s

        result = sympify(func).subs(ssym, s / scale) / abs(scale)

        if shift != 0:
            result = result * sym.exp(s * shift / scale)
        return result

    def integral(self, expr, t, s):

        const, expr = factor_const(expr, t)

        if len(expr.args) != 2:
            self.error('Expecting two args')

        integrand = expr.args[0]

        if not isinstance(expr, sym.Integral):
            self.error('Expecting integral')

        if len(expr.args[1]) != 3:
            self.error('Require definite integral')

        var = expr.args[1][0]
        limits = expr.args[1][1:]
        const2, expr2 = factor_const(integrand, var)

        if (expr2.is_Function and
            expr2.args[0] == t - var and limits[0] == 0 and limits[1] == sym.oo):
            return const2 * self.term(expr2.subs(t - var, t), t, s) / s

        # Handle integral(x(tau), (tau, -oo, t))
        if not limits[0].is_positive and limits[1] == t:
            if isinstance(expr2, AppliedUndef):
                if expr2.args[0] == expr.args[1][0]:
                    return const2 * self.func(expr2, expr2.args[0], s) / s

        # Look for convolution integral
        if limits[0].is_positive:
            self.error('Cannot handle lower limit %s' % limits[0])

        if limits[1] < t:
            self.error('Cannot handle upper limit %s' % limits[1])

        if ((len(expr.args) != 2) or not expr2.is_Mul or
            not expr2.args[0].is_Function or not expr2.args[1].is_Function):
            self.error('Need integral of product of two functions')

        f1 = expr2.args[0]
        f2 = expr2.args[1]
        # TODO: apply similarity theorem if have f(a * tau) etc.

        if (f1.args[0] == var and f2.args[0] == t - var):
            F1 = self.term(f1, var, s)
            F2 = self.term(f2.subs(t - var, t), t, s)
        elif (f2.args[0] == var and f1.args[0] == t - var):
            F1 = self.term(f1.subs(t - var, t), t, s)
            F2 = self.term(f2, var, s)
        else:
            self.error('Cannot recognise convolution')

        return const2 * F1 * F2

    def derivative_undef(self, expr, t, s):

        if not isinstance(expr, sym.Derivative):
            self.error('Expecting derivative')

        if (not isinstance(expr.args[0], AppliedUndef) and
            expr.args[1][0] != t):
            self.error('Expecting function of t')

        ssym = sympify(str(s))
        name = expr.args[0].func.__name__
        func1 = name[0].upper() + name[1:] + '(%s)' % str(ssym)
        return sympify(func1).subs(ssym, s) * s ** expr.args[1][1]

    def sin_cos(self, expr, t, s):

        # Handle exp(-alpha * t) * sin(omega * t + phi) * u(t - tau)
        # The exp(-alpha * t) and u(t - tau) parts are optional.

        # Sympy sometimes has problems with this...

        factors = expr.as_ordered_factors()

        if len(factors) > 3:
            raise ValueError('Not expsin, too many factors')

        alpha = 0
        beta = 0
        m = 0
        if (factors[m].is_Function and factors[m].func is sym.exp):
            exparg = factors[m].args[0]
            alpha, beta = scale_shift(exparg, t)
            m += 1

        if not (factors[m].is_Function and factors[m].func in (sym.sin, sym.cos)):
            raise ValueError('Not expsin, no sin/cos')

        sincosarg = factors[m].args[0]
        omega, phi = scale_shift(sincosarg, t)

        if factors[m].func is sym.cos:
            phi += sym.pi / 2

        m += 1

        if len(factors) == m + 1 and not (factors[m].is_Function and factors[m].func is sym.Heaviside):
            raise ValueError('Not expsin, no Heaviside')

        tau = 0
        if len(factors) == m + 1:
            eta, zeta = scale_shift(factors[m].args[0], t)
            if eta != 1:
                raise ValueError('Need to use similarity theorem')
            tau = -zeta

            if tau.is_negative:
                tau = 0
            elif tau.is_Symbol and not tau.is_positive:
                print('Assuming %s is positive' % tau)
            elif (tau.is_Mul and len(tau.args) == 2 and
                  tau.args[0].is_negative and tau.args[1].is_Symbol):
                print('Assuming %s is positive' % -tau)
                tau = 0

        if tau != 0:
            phi += omega * tau

        E = (omega * sym.cos(phi) + (s - alpha) * sym.sin(phi)) / (omega**2 + (s - alpha)**2)

        if tau != 0:
            E *= sym.exp(-tau * s)
            if alpha != 0:
                E *= sym.exp(alpha * tau)

        if beta != 0:
            E = sym.exp(beta) * E

        return E

    def term(self, expr, t, s):

        # Unilateral LT ignores expr for t < 0 so remove Piecewise.
        if expr.is_Piecewise and expr.args[0].args[1].has(t >= 0):
            expr = expr.args[0].args[0]

        const, expr = factor_const(expr, t)

        terms = expr.expand(deep=False).as_ordered_terms()
        if len(terms) > 1:
            result = 0
            for term in terms:
                result += self.term(term, t, s)
            return const * result

        tsym = sympify(str(t))
        expr = expr.replace(tsym, t)

        if expr.has(sym.Integral):
            return self.integral(expr, t, s) * const

        if expr.has(sym.sin, sym.cos):
            try:
                return self.sin_cos(expr, t, s) * const
            except:
                pass

        if expr.has(AppliedUndef):

            if expr.has(sym.Derivative):
                return self.derivative_undef(expr, t, s) * const

            factors = expr.as_ordered_factors()
            if len(factors) == 1:
                return const * self.func(factors[0], t, s)
            elif len(factors) > 2:
                self.error('Cannot handle product')

            foo = factors[1]
            if foo.is_Function and foo.func == sym.exp and foo.args[0].has(t):
                scale, shift = scale_shift(foo.args[0], t)
                if shift == 0:
                    result = self.func(factors[0], t, s)
                    return const * result.subs(s, s - scale)
            self.error('Cannot handle product')

        if expr.has(sym.Heaviside(t)):
            return self.do_0(expr.replace(sym.Heaviside(t), 1), t, s) * const

        if expr.has(sym.DiracDelta) or expr.has(sym.Heaviside):
            try:
                return self.do_0minus(expr, t, s) * const
            except ValueError:
                pass

        return self.do_0(expr, t, s) * const


laplace_transformer = LaplaceTransformer()


def laplace_transform(expr, t, s, evaluate=True, **kwargs):
    """Compute unilateral Laplace transform of expr with lower limit 0-.
    """

    return laplace_transformer.transform(expr, t, s,
                                         evaluate=evaluate,
                                         **kwargs)


def LT(expr, t, s, evaluate=True, **kwargs):
    """Compute unilateral Laplace transform of expr with lower limit 0-.
    """

    return laplace_transformer.transform(expr, t, s,
                                         evaluate=evaluate,
                                         **kwargs)



"""
This module performs schematic drawing using circuitikz from a netlist.

>>> from lcapy import Schematic
>>> sch = Schematic()
>>> sch.add('P1 1 0.1; down')
>>> sch.add('R1 3 1; right')
>>> sch.add('L1 2 3; right')
>>> sch.add('C1 3 0; down')
>>> sch.add('P2 2 0.2; down')
>>> sch.add('W 0 0.1; right')
>>> sch.add('W 0.2 0.2; right')
>>> sch.draw()

Copyright 2014--2021 Michael Hayes, UCECE
"""

# Strings starting with ;; are schematic options.  They are parsed in
# netfile.py and added to the opts attribute of the netlist.  They get
# passed to the draw method of the Schematic class.


from __future__ import print_function
from .latex import latex_format_label, latex_format_node_label
from .expr import Expr
from . import schemcpts
import sympy as sym
from .engformatter import EngFormatter
from .schemmisc import Pos
from .schemplacer import schemplacer
from .opts import Opts
from .netfile import NetfileMixin
from .system import LatexRunner, PDFConverter, tmpfilename
from os import path, remove
from collections import OrderedDict
from warnings import warn
import math


__all__ = ('Schematic', )


def display_matplotlib(filename, dpi=150, scale=2):

    from matplotlib.pyplot import figure, figaspect
    from matplotlib.image import imread

    img = imread(filename)
    h, w, d = img.shape
    width = scale * w / dpi
    height = scale * h / dpi
    fig = figure(figsize=(width, height))
    ax = fig.add_subplot(111)
    ax.imshow(img)
    ax.axis('off')
    fig.subplots_adjust(left=0.05, right=0.95, top=0.95, bottom=0.05)


def png_image_size(filename):

    import struct

    with open(filename, 'rb') as f:
        header = f.read(25)

    if (header[:8] != b'\211PNG\r\n\032\n' and (header[12:16] != b'IHDR')):
        raise Exception('%s not a png image' % filename)
    w, h = struct.unpack('>LL', header[16:24])
    width = int(w)
    height = int(h)
    return width, height


class SchematicOpts(Opts):

    def __init__(self):

        super (SchematicOpts, self).__init__(
            {'draw_nodes': 'primary',
             'label_values': True,
             'label_ids': True,
             'annotate_values': False,
             'label_nodes': 'primary',
             'scale' : 1.0,
             'dpi' : 150,
             'cpt_size' : 1.5,
             'node_spacing' : 2.0,
             'help_lines' : 0.0,
             'style' : 'american',
             'voltage dir' : 'RP'})


class Node(object):

    def __init__(self, name):

        self.name = name
        self._port = False
        self._count = 0
        parts = name.split('_')
        # Primary 1, 2, a, a_3.  Not primary _1, _2, _a, _a_3, 0_3, a_b_c
        self.primary = (name[0] != '_' and len(parts) <= 2) and not (name[0].isdigit() and len(parts) != 1)

        # See MX1.sch for example.
        if '._' in name:
            self.primary = False

        self.elt_list = []
        self.pos = 'unknown'
        # Sanitised name
        self.s = name.replace('.', '@')
        self.label = latex_format_node_label(self.name)
        self.labelpos = None
        self.pin = False
        self.pinlabel = ''
        self.namepos = None
        self.pinname = ''
        self.pinpos = None
        self.clock = False
        self.auxiliary = None
        # Reference to node that defines the pin
        self.ref = None
        # Node is connected to a ground, etc. (an implicit power
        # supply connection)
        self.implicit = False

    @property
    def basename(self):
        fields = self.name.split('.')
        return fields[-1]

    @property
    def cptname(self):
        fields = self.name.split('.')
        if len(fields) < 2:
            return None
        return fields[-2]

    def __repr__(self):
        return '%s @ (%s)' % (self.name, self.pos)

    def append(self, elt):
        """Add new element to the node"""

        if elt.type == 'P':
            self._port = True

        self.elt_list.append(elt)
        if elt.type not in ('A', 'O'):
            self._count += 1

    @property
    def count(self):
        """Number of elements (including wires but not open-circuits)
        connected to the node"""

        return self._count

    def belongs(self, cpt_name):
        return self.cptname == cpt_name

    def visible(self, draw_nodes):
        """Return true if node drawn.
        `draw_nodes' can be `all', 'none', 'connections', 'primary', None,
        True, or False."""

        if self.auxiliary:
            return False

        if draw_nodes in ('all', True):
            return True

        if self.pin:
            return False

        if self.implicit:
            return False

        if self._port:
            return True

        if draw_nodes in ('none', None, False):
            return False

        # Implied port
        if self.count == 1:
            return True

        if draw_nodes in ('connections', 'connected'):
            return self.count > 2

        if draw_nodes == 'primary':
            return self.primary

        raise ValueError('Unknown argument %s for draw_nodes' % draw_nodes)

    @property
    def port(self):
        """Return true if node is a port"""

        return self._port or self.count == 1

    def show_label(self, label_nodes):

        if self.label == '':
            return False

        name = self.basename

        # pins is for backward compatibility
        if label_nodes in ('none', 'pins', 'false', False):
            return False
        elif label_nodes in ('all', 'true', True):
            return True
        elif label_nodes == 'alpha':
            return self.primary and name[0].isalpha()
        elif label_nodes == 'primary':
            return self.primary

        # handle label_nodes = '{1, 2}' etc.
        if label_nodes[0] == '{' and label_nodes[-1] == '}':
            label_nodes = label_nodes[1:-1]
        labels = [foo.strip() for foo in label_nodes.split(',')]
        return self.label in labels

    def debug(self):
        print(' %s @ (%s), count=%d, pin=%s' % (self.name, self.pos,
                                                self.count, self.pin))


class Schematic(NetfileMixin):

    def __init__(self, filename=None, allow_anon=False, **kwargs):

        self.elements = OrderedDict()
        self.nodes = {}
        self.hints = False
        self._init_parser(schemcpts, allow_anon)
        self.cpt_size = 1.2
        self.node_spacing = 2.0
        self.scale = 1.0
        self.dummy_node = 0
        self.context = None
        self.debug = 0

        if filename is not None:
            self.netfile_add(filename)

    def __repr__(self):

        return self.netlist()

    def __getitem__(self, name):
        """Return component by name"""
        try:
            return self.elements[name]
        except KeyError:
            raise AttributeError('Unknown component %s' % name)

    def netfile_add(self, filename):
        """Add the nets from file with specified filename"""

        self._netfile_add(filename)

    def add(self, string):
        """Add a component to the netlist.
        The general form is: 'Name Np Nm args'
        where Np is the positive node and Nm is the negative node.

        A positive current is defined to flow from the positive node
        to the negative node.
        """

        self._add(string)

    def netlist(self):
        """Return the current netlist"""

        return '\n'.join([elt.__str__() for elt in self.elements.values()])

    def _node_add(self, nodename, elt, auxiliary=False):

        if nodename not in self.nodes:
            self.nodes[nodename] = Node(nodename)
        node = self.nodes[nodename]

        node.append(elt)

        if node.auxiliary is None:
            node.auxiliary = auxiliary
        elif node.auxiliary and not auxiliary:
            # An auxiliary node can be made a proper node.
            node.auxiliary = False

        return node

    def _cpt_add(self, cpt):

        if cpt.offset != 0 and len(cpt.node_names) == 2:
            # Change component and add wires.

            n1 = cpt.node_names[0]
            n2 = cpt.node_names[1]
            # If have only two parallel components, can used fixed node names.
            self.dummy_node += 1
            on1 = n1 + 'off%d' % self.dummy_node
            self.dummy_node += 1
            on2 = n2 + 'off%d' % self.dummy_node

            angle = 90
            size = cpt.offset
            if size < 0:
                size = -size
                angle = -angle

            w1 = 'W %s %s; rotate=%s, size=%s' % (n1, on1, cpt.angle + angle, size)
            w2 = 'W %s %s; rotate=%s, size=%s' % (n2, on2, cpt.angle + angle, size)

            self.add(w1)
            self.add(w2)

            # Add open-circuit to ensure alignment.
            o = 'O %s %s; rotate=%s, size=%s' % (n1, n2, cpt.angle, cpt.size)
            self.add(o)

            # Rename nodes
            parts = cpt.net.split(' ')
            parts[1] = on1
            parts[2] = on2
            net = ' '.join(parts)
            cpt.opts.strip('offset')
            self.add('%s; %s' % (net, cpt.opts))
            return

        def tex_name(name, subscript=None):

            if subscript is None:
                subscript = ''

            if len(name) > 1:
                name = r'\mathrm{%s}' % name
            if len(subscript) > 1:
                subscript = r'\mathrm{%s}' % subscript
            if len(subscript) == 0:
                return name

            return '%s_{%s}' % (name, subscript)

        # There are two possible labels for a component:
        # 1. Component identifier, e.g., R1
        # 2. Component value, expression, or symbol
        id_label = tex_name(cpt.type, cpt.id)
        value_label = None

        if cpt.type in ('O', 'P', 'W') or id_label.find('#') != -1:
            id_label = None

        if cpt.type in ('S', 'SW', 'U'):
            value_label = ''

        if cpt.args != ():

            # TODO, extend for mechanical and acoustical components.
            units_map = {'V': 'V', 'I': 'A', 'R': '$\Omega$',
                         'C': 'F', 'L': 'H'}

            expr = cpt.args[0]
            if cpt.classname in ('Vstep', 'Istep'):
                expr = '(%s) * Heaviside(t)' % expr
                value_label = Expr(expr, cache=False).latex_math()
            elif cpt.classname in ('Vs', 'Is'):
                value_label = Expr(expr, cache=False).latex_math()
            elif cpt.classname == 'TF':
                expr = sym.sympify(expr)
                if expr.is_Pow and expr.args[1] == -1:
                    value_label = '%s:1' % (1 / expr)
                else:
                    value_label = '1:%s' % expr
            elif cpt.type in ('F', 'H') and len(cpt.args) > 1:
                # This is hard to give a reasonable label since the
                # control current is specified by a voltage source.
                # The user will have to override manually.
                expr = cpt.args[1]
                value_label = Expr(expr, cache=False).latex_math()
            elif cpt.classname not in ('TP',):
                try:
                    # Handle things like 9/1000 that can occur
                    # when substituting cpt values.
                    value = float(sym.Rational(expr))
                    if cpt.type in units_map:
                        value_label = EngFormatter().latex_math(value, units_map[cpt.type])
                    else:
                        value_label = Expr(expr, cache=False).latex_math()

                except (ValueError, TypeError):
                    # This catches non numeric arg.
                    value_label = Expr(expr, cache=False).latex_math()

        # Currently, we only annnotated the component with the value,
        # expression, or symbol.  If this is not specified, it
        # defaults to the component identifier.  Note, some objects
        # we do not want to label, such as wires and ports.
        cpt.id_label = '' if id_label is None else latex_format_label(id_label)
        cpt.value_label = cpt.id_label if value_label is None else latex_format_label(value_label)

        if cpt.opts_string != '':
            self.hints = True

        if cpt.name in self.elements:
            warn('Overriding component %s' % cpt.name)
            # Need to search lists and update component.

        self.elements[cpt.name] = cpt

        for node in cpt.auxiliary_node_names:
            self._node_add(node, cpt, auxiliary=True)

        # Note, an auxiliary node can be trumped...
        for node in cpt.required_node_names:
            self._node_add(node, cpt, auxiliary=False)

    def _setup(self):
        # This is called before node positions are assigned.

        for elt in self.elements.values():
            elt.setup()

        for nodename, node in self.nodes.items():
            if node.ref is not None:
                continue
            if node.cptname is not None and not node.implicit:
                raise ValueError('Unreferenced pin connection %s for %s' % (node.name, node.elt_list))

    def make_graphs(self, debug=0):

        placer = schemplacer(self.elements, self.nodes, 'graph', debug)
        placer._make_graphs()
        return placer.xgraph, placer.ygraph

    def _positions_calculate(self, method='graph', debug=False):

        placer = schemplacer(self.elements, self.nodes, method, debug)
        self.width, self.height = placer.solve(self.node_spacing)

    def _tikz_draw(self, style_args='', **kwargs):

        method = kwargs.pop('method', 'graph')

        self._setup()

        self._positions_calculate(method, self.debug)

        # Note, scale does not scale the font size.
        opts = ['scale=%.2f' % self.scale,
                'transform shape',
                '/tikz/circuitikz/bipoles/length=%.2fcm' % self.cpt_size]
        opts.append(style_args)
        if 'options' in kwargs:
            opts.append(kwargs.pop('options'))

        if 'font' in kwargs:
            font = kwargs.pop('font')
            opts.append('font=' + font)

        if 'voltage dir' in kwargs:
            voltage_dir = kwargs.pop('voltage dir')
            opts.append('voltage dir=' + voltage_dir)

        s = r'\begin{tikzpicture}[%s]''\n' % ', '.join(opts)

        # Add preamble
        if 'preamble' in kwargs:
            s += '  ' + kwargs.pop('preamble') + '\n'

        help_lines = float(kwargs.pop('help_lines', 0))
        color = kwargs.pop('color', 'blue')
        if help_lines != 0:
            start = Pos(-0.5, -0.5) * self.node_spacing
            stop = Pos(self.width + 0.5, self.height + 0.5) * self.node_spacing

            s += r'  \draw[help lines, %s] (%s) grid [xstep=%s, ystep=%s] (%s);''\n' % (
                color, start, help_lines, help_lines, stop)

        # Write coordinates.  TODO, not all coordinates are needed
        # so those can be weeded out to simplify the generated file.
        for n in self.nodes.values():
            s += r'  \coordinate (%s) at (%s);''\n' % (n.s, n.pos)

        # Keyword args for second pass
        kwargs2 = kwargs.copy()

        # Pass 1: Draw components
        for elt in self.elements.values():
            if elt.ignore:
                continue
            if elt.directive:
                for key, val in elt.opts.items():
                    # Update kwargs
                    kwargs[key] = val

            s += elt.draw(**kwargs)
            s += elt.draw_nodes(**kwargs)
            s += elt.draw_pins()

        # Pass 2: Add the node labels
        for elt in self.elements.values():
            if elt.ignore:
                continue
            if elt.directive:
                for key, val in elt.opts.items():
                    kwargs2[key] = val
            s += elt.draw_node_labels(**kwargs2)

        # Add postamble
        if 'postamble' in kwargs:
            s += '  ' + kwargs.pop('postamble') + '\n'

        s += r'\end{tikzpicture}''\n'

        return s

    def tikz_draw(self, filename, **kwargs):

        root, ext = path.splitext(filename)

        self.debug = kwargs.pop('debug', 0)
        style = kwargs.pop('style', 'american')
        self.dpi = float(kwargs.pop('dpi', 150))
        self.cpt_size = float(kwargs.pop('cpt_size', 1.2))
        self.node_spacing = float(kwargs.pop('node_spacing', 2.0))
        self.scale = float(kwargs.pop('scale', 1.0))

        include = ''
        if 'include' in kwargs:
            include += kwargs.pop('include')
        if 'includefile' in kwargs:
            file = open(kwargs.pop('includefile'))
            include += file.read()
            file.close()

        if style == 'american':
            style_args = 'american currents, american voltages'
        elif style == 'british':
            style_args = 'american currents, european voltages'
        elif style == 'european':
            style_args = ('european currents, european voltages'
                          ', european inductors, european resistors')
        else:
            raise ValueError('Unknown style %s' % style)

        # For debugging when do not want to write to file
        nosave = kwargs.pop('nosave', False)

        latexrunner = LatexRunner(self.debug & 2)

        self.circuitikz_date, self.circuitikz_version = latexrunner.circuitikz_version()
        if self.circuitikz_date is None:
            raise RuntimeError('circuitikz is not installed')

        content = self._tikz_draw(style_args=style_args, **kwargs)

        if nosave:
            return

        if self.debug & 1:
            print('circuitikz version %s (%s)' % (self.circuitikz_version,
                                                  self.circuitikz_date))
            print('width=%d, height=%d, dpi=%d, cpt_size=%.2f, node_spacing=%.2f, scale=%.2f'
                  % (self.width, self.height, self.dpi,
                     self.cpt_size, self.node_spacing, self.scale))
            print('Nodes:')
            for node in self.nodes.values():
                node.debug()

        if ext in ('.pytex', '.schtex', '.pgf'):
            if include != '':
                warn('Include option ignored for non-standalone file')
            open(filename, 'w').write(content)
            return

        # Need amsmath for operatorname
        template = ('\\documentclass[a4paper]{standalone}\n'
                    '\\usepackage{amsmath}\n'
                    '\\usepackage{circuitikz}\n'
                    '\\usetikzlibrary{fit, shapes, arrows, patterns, decorations.text, decorations.markings}\n'
                    '%s\n'
                    '\\begin{document}\n%s\\end{document}')
        content = template % (include, content)
        tex_filename = filename.replace(ext, '.tex')
        open(tex_filename, 'w').write(content)

        if ext == '.tex':
            return

        pdf_filename = tex_filename.replace('.tex', '.pdf')
        latexrunner.run(tex_filename)
        if not (self.debug & 1):
            latexrunner.cleanup(tex_filename, pdf_filename)

        if not path.exists(pdf_filename):
            raise RuntimeError('Could not generate %s with pdflatex' %
                               pdf_filename)

        if ext == '.pdf':
            return

        if ext == '.svg':
            pdfconverter = PDFConverter(self.debug & 2)
            pdfconverter.to_svg(pdf_filename, root + '.svg')
            if not (self.debug & 1):
                remove(pdf_filename)
            return

        if ext == '.png':
            pdfconverter = PDFConverter(self.debug & 2)
            pdfconverter.to_png(pdf_filename, root + '.png', self.dpi)
            if not (self.debug & 1):
                remove(pdf_filename)
            return

        raise RuntimeError('Cannot create file of type %s' % ext)

    def draw(self, filename=None, **kwargs):
        """
        filename specifies the name of the file to produce.  If None,
        the schematic is displayed on the screen.

        Note, if using Jupyter, then need to first issue command %matplotlib inline

        kwargs include:
           'label_ids': True to show component ids
           'label_values': True to display component values
           'annotate_values': True to display component values as separate label
           'draw_nodes': True to show all nodes,
             False or 'none' to show no nodes,
             'primary' to show primary nodes,
             'connections' to show nodes that connect more than two components,
             'all' to show all nodes
           'label_nodes': True to label all nodes,
             False or 'none' to label no nodes,
             'primary' to label primary nodes,
             'alpha' to label nodes starting with a letter,
             'pins' to label nodes that are pins on a chip,
             'all' to label all nodes,
             'none' to label no nodes
           'include': name of file to include before \\begin{document}
           'style': 'american', 'british', or 'european'
           'scale': schematic scale factor, default 1.0
           'node_spacing': spacing between component nodes, default 2.0
           'cpt_size': size of a component, default 1.5
           'dpi': dots per inch for png files
           'help_lines': distance between lines in grid, default 0.0 (disabled)
           'debug': Non-zero to display debug information
        """

        # None means don't care, so remove.
        for key in list(kwargs.keys()):
            if kwargs[key] is None:
                kwargs.pop(key)

        # Remove options that may be overridden
        for elt in self.elements.values():
            for key in list(elt.opts.keys()):
                if key in kwargs:
                    elt.opts.remove(key)

        # Default options
        opts = SchematicOpts()
        for key, val in opts.items():
            if key not in kwargs:
                kwargs[key] = val

        # Global options (usually at end of netlist for historical reasons)
        # These can be overwritten
        for eltname in self.elements:
            elt = self.elements[eltname]
            if not elt.directive:
                continue
            for key, val in elt.opts.items():
                # val is a str
                kwargs[key] = val

        def in_ipynb():
            try:
                ip = get_ipython()
                cfg = ip.config

                kernapp = cfg['IPKernelApp']

                # Check if processing ipynb file for Jupyter notebook.
                if 'connection_file' in kernapp:
                    return True
                elif kernapp and kernapp['parent_appname'] == 'ipython-notebook':
                    return True
                else:
                    return False
            except (NameError, KeyError):
                return False

        if not self.hints:
            raise RuntimeWarning('No schematic drawing hints provided!')

        if in_ipynb() and filename is None:
            png = 'png' in kwargs and kwargs.pop('png')
            svg = 'svg' in kwargs and kwargs.pop('svg')

            if not png and not svg:
                svg = False

            if svg:
                try:
                    from IPython.display import SVG, display_svg

                    svgfilename = tmpfilename('.svg')
                    self.tikz_draw(svgfilename, **kwargs)

                    # Create and display SVG image object.
                    # Note, there is a problem displaying multiple SVG
                    # files since the later ones inherit the namespace of
                    # the first ones.
                    display_svg(SVG(filename=svgfilename))
                    return
                except:
                    pass

            from IPython.display import Image, display_png

            png_filename = tmpfilename('.png')
            self.tikz_draw(png_filename, **kwargs)

            # Create and display PNG image object.
            # There are two problems:
            # 1. The image metadata (width, height) is ignored
            #    when the ipynb file is loaded.
            # 2. The image metadata (width, height) is not stored
            #    when the ipynb file is written non-interactively.
            width, height = png_image_size(png_filename)
            # width, height specify the image dimension in pixels
            display_png(Image(filename=png_filename,
                              width=width, height=height))
            if not (self.debug & 1):
                remove(png_filename)
            return

        if filename is None:
            filename = tmpfilename('.png')
            # Thicken up lines to reduce aliasing causing them to
            # disappear, especially when using pdftoppm.
            self.tikz_draw(filename=filename,
                           options='/tikz/circuitikz/bipoles/thickness=2',
                           **kwargs)
            display_matplotlib(filename, self.dpi)
            if not (self.debug & 1):
                remove(filename)
            return

        self.tikz_draw(filename=filename, **kwargs)

def test():

    sch = Schematic()

    sch.add('P1 1 0.1')
    sch.add('R1 1 3; right')
    sch.add('L1 3 2; right')
    sch.add('C1 3 0; down')
    sch.add('P2 2 0.2')
    sch.add('W 0.1 0; right')
    sch.add('W 0 0.2; right')

    sch.draw()
    return sch



from lcapy import delta
from lcapy.discretetime import n, dt
from matplotlib.pyplot import savefig

x = delta(n) + delta(n - 2)
abs(x.DTFT().subs(dt, 1)).plot(norm=True, figsize=(6, 3))

savefig('dt1-DTFT-plot1.png', bbox_inches='tight')



from .expr import expr
from .immittancemixin import ImmittanceMixin
from .quantity import Quantity

class ImpedanceMixin(Quantity, ImmittanceMixin):

    quantity = 'impedance'
    quantity_label = 'Impedance'
    quantity_units = 'ohm'
    is_impedance = True
    is_immittance = True
    is_ratio = True

    # Immittances derived from a realisable circuit will be causal but
    # non-causal immittances can also be constructed.  So this might
    # disappear.  An example non-causal impulse response is z(t) = R delta(t + T)
    # with an impedance Z(s) = R exp(s * T).  This has
    # a real part R * exp(re(s) * T) * cos(T * im(s))
    # and imaginary part R * exp(re(s) * T) * sin(T * im(s))
    
    is_always_causal = True
    

    @property    
    def Y(self):
        """Admittance."""

        from .admittance import admittance

        return admittance(1 / self)

    @property
    def Z(self):
        """Impedance."""
        return self
    
    def __rtruediv__(self, x):
        """Reverse true divide"""

        x = expr(x)
        if x.is_constant:
            from .admittance import admittance            
            return admittance(x.expr / self.expr)
        return super(ImpedanceMixin, self).__rtruediv__(x)
    
    def cpt(self):
        from .oneport import R, C, L, Z
        from .symbols import s        

        if self.is_number or self.is_dc:
            return R(self.expr)

        z = self.laplace() * s

        if z.is_number:
            return C((1 / z).expr)

        z = self.laplace() / s

        if z.is_number:
            return L(z.expr)

        return Z(self)




"""This module performs transformations between domains.

Copyright 2018--2021 Michael Hayes, UCECE

"""

from .sym import sympify, pi
from .sym import dt
from .symbols import f, s, t, omega, j, jw, jw0, Omega, F, k, n, z
from .symbols import domain_vars, domain_var_ids
from .nexpr import n
from .expr import expr as expr1
from .expr import Expr
from numpy import ndarray, array


def transform(expr, arg, **assumptions):
    """If arg is a domain variable perform domain transformation,
    otherwise perform substitution.

    Note (1 / s)(omega) will fail since 1 / s is assumed not to be
    causal and so the Fourier transform is unknown.  However,
    impedance(1 / s)(omega) will work since an impedance is assumed to
    be causal.  Alternatively, use (1 / s)(omega, causal=True).

    Transforming from s->jomega is fast since it just requires
    a substitution of s with jomega.

    Transforming from s->omega, s->Omega, s->f, or s->F can be slow since
    this requires a inverse Laplace transform followed by a Fourier
    transform.  However, if the expression is causal and the
    expression is lossy when s is replaced by jw, the result can be
    found by substituting jw or 2 * 2 * pi * f for s.  This does not
    apply for an expression such as Z = 1 / (s * C).

    """

    arg = expr1(arg)

    # Handle things like (3 * s)(5 * s)
    if isinstance(expr, arg.__class__) and not isinstance(expr, Superposition):
        return expr.subs(arg)

    # Handle expr(t), expr(s), expr(f), expr(omega), expr(jomega), expr(3j)
    try:
        if arg is t:
            return expr.time(**assumptions)
        elif arg is s:
            return expr.laplace(**assumptions)
        elif arg is f:
            return expr.fourier(**assumptions)
        elif arg is omega:
            return expr.angular_fourier(**assumptions)
        elif arg is Omega:
            return expr.norm_angular_fourier(**assumptions)
        elif arg is F:
            return expr.norm_fourier(**assumptions)
        elif arg.has(j):
            return expr.phasor(omega=arg / j, **assumptions)
        elif arg is n:
            return expr.discrete_time(**assumptions)
        elif arg is k:
            return expr.discrete_frequency(**assumptions)
        elif arg is z:
            return expr.zdomain(**assumptions)
    except AttributeError:
        raise AttributeError('Transform of %s domain to %s domain not implemented' % (expr.domain, arg.domain))

    # Handle expr(texpr), expr(sexpr), expr(fexpr), expr(omegaexpr).
    # For example, expr(2 * f).
    result = None
    if isinstance(arg, TimeDomainExpression):
        result = expr.time(**assumptions)
    elif isinstance(arg, LaplaceDomainExpression):
        result = expr.laplace(**assumptions)
    elif isinstance(arg, FourierDomainExpression):
        result = expr.fourier(**assumptions)
    elif isinstance(arg, NormFourierDomainExpression):
        result = expr.norm_fourier(**assumptions)
    elif isinstance(arg, AngularFourierDomainExpression):
        result = expr.angular_fourier(**assumptions)
    elif isinstance(arg, NormAngularFourierDomainExpression):
        result = expr.norm_angular_fourier(**assumptions)
    elif arg.has(j):
        result = expr.phasor(omega=arg / j, **assumptions)
    elif isinstance(arg, DiscreteTimeDomainExpression):
        result = expr.discrete_time(**assumptions)
    elif isinstance(arg, DiscreteFourierDomainExpression):
        result = expr.discrete_frequency(**assumptions)
    elif isinstance(arg, ZDomainExpression):
        result = expr.zdomain(**assumptions)
    elif arg.is_constant:
        if not isinstance(expr, Superposition):
            result = expr.time(**assumptions)
        else:
            result = expr
    else:
        raise ValueError('Can only return t, f, s, F, omega, Omega, or jw domains')

    return result.subs(arg, **assumptions)


def call(expr, arg, **assumptions):

    if isinstance(arg, (tuple, list)):
        return [expr._subs1(expr.var, arg1) for arg1 in arg]

    elif isinstance(arg, ndarray):
        return array([expr._subs1(expr.var, arg1) for arg1 in arg])

    arg = expr1(arg)
    if arg.is_constant:
        return expr.subs(arg)

    return expr.transform(arg, **assumptions)


def select(expr, kind):

    if not isinstance(kind, str):
        return expr.subs(j * kind)

    # If kind is an expr, then will add 'dc', 'time', etc. as symbols.

    if kind == 't':
        return expr.time()
    elif kind in ('dc', 'time'):
        return expr.subs(0)
    elif kind in ('s', 'ivp', 'super', 'laplace'):
        return expr.laplace()
    elif kind == 'f':
        return expr.fourier()
    elif kind == 'omega':
        return expr.angular_fourier()
    elif kind == 'Omega':
        return expr.norm_angular_fourier()
    elif kind == 'F':
        return expr.norm_fourier()
    elif isinstance(kind, str) and kind.startswith('n'):
        return expr.angular_fourier()
    else:
        raise RuntimeError('unknown kind')


from .fexpr import FourierDomainExpression
from .sexpr import LaplaceDomainExpression
from .texpr import TimeDomainExpression
from .omegaexpr import AngularFourierDomainExpression
from .normfexpr import NormFourierDomainExpression
from .normomegaexpr import NormAngularFourierDomainExpression
from .nexpr import DiscreteTimeDomainExpression
from .kexpr import DiscreteFourierDomainExpression
from .zexpr import ZDomainExpression
from .super import Superposition
from .current import current
from .voltage import voltage
from .admittance import admittance
from .impedance import impedance
from .transfer import transfer



"""This module provides the State class that maintains the global state,
such as the symbol context.

Copyright 2014--2021 Michael Hayes, UCECE

"""

from sympy.assumptions.assume import global_assumptions

from .context import Context
from .config import check_units, abbreviate_units, loose_units, show_units
from .config import canonical_units, printing_order
from .printing_config import PrintingConfig
from copy import copy

class State(object):
    """This maintains Lcapy's state, such as the defined symbols, and
    behaviour.

    `loose_units` (default True) allows constants to be added to quantities
    `show_units` (default False) prints the units after an expression
    `canonical_units` (default True) converts units to canonical form, e.g., V / A is
    shown as ohms.
    'printing.abbreviate_units` (default True) prints V rather than volts
    'printing.order` (default `none`) controls the order that symbols in an expression are printed.

    """

    def __init__(self):
        self.global_context = Context()
        self.context = self.global_context
        self.previous_context = []
        self.printing = PrintingConfig()
        
        # With loose_units can add constants to quantities, e.g., voltage(1) + 2
        # or impedance(2) == 2.
        self.loose_units = loose_units
        self.check_units = check_units
        self.show_units = show_units
        self.canonical_units = canonical_units
        self.printing.abbreviate_units = abbreviate_units
        self.printing.order = printing_order

    def new_context(self):

        context = Context()
        context.symbols = copy(self.global_context.symbols)
        return context
        
    def switch_context(self, context):

        self.previous_context.append(self.context)
        self.context = context

        global_assumptions.clear()
        global_assumptions.update(self.context.assumptions)

    def restore_context(self):

        self.context.assumptions.update(global_assumptions)
        
        self.context = self.previous_context.pop()
        
        global_assumptions.clear()
        global_assumptions.update(self.context.assumptions)

    @property
    def abbreviate_units(self):

        return self.printing.abbreviate_units

    @abbreviate_units.setter
    def abbreviate_units(self, val):
        """Enable/disable printing of units in abbreviated form."""

        self.printing.abbreviate_units = val


state = State()
